;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; PROGRAMMING FUNDAMENTALS I - 2004
; MICHELE LANZA - FACULTY OF INFORMATICS - UNIVERSITY OF LUGANO, SWITZERLAND
; LECTURE 4 - COMPOUND DATA II AND SYNTAX & SEMANTICS
; QUADRATIC EQUATION EXAMPLE

; DATA DEFINITION AND ANALYSIS
(define-struct equation (a b c))
; a (quadratic) equation is a structure: (make-equation a b c) where
; a, b, and c are numbers
; EXAMPLES
(define eq1 (make-equation 1 0 -1))
(define eq2 (make-equation 2 4 2))
(define eq3 (make-equation 1 1 1))
(define eq4 (make-equation 0 2 3))
; TEMPLATE
; (define foo-equation
; lambda (eq)
; ..(equation-a eq)..
; ..(equation-b eq)..
; ..(equation-c eq)..
; ..eq..

; helper function discriminant: equation -> number
; takes a quadratic equation and computes its discriminant
(define discriminant
  (lambda (eq)
    (- (sqr (equation-b eq)) (* 4 (equation-a eq) (equation-c eq)))))

; how-many-solutions2: equation -> number or string
; consumes a quadratic equation and produces the number of
; solutions, if any
(define how-many-solutions2
  (lambda (eq)
    (cond
      [(not (struct? eq)) (error 'how-many-solutions "expected an equation as input")]
      [else
       (cond
         [(= 0 (equation-a eq)) "degenerate"]
         [(> (sqr (equation-b eq)) (* 4 (equation-a eq) (equation-c eq))) 2]
         [(= (sqr (equation-b eq)) (* 4 (equation-a eq) (equation-c eq))) 1]
         [else 0])])))

; how-many-solutions: equation -> number or string
; consumes a quadratic equation and produces the number of
; solutions, if any
(define how-many-solutions
  (lambda (eq)
    (cond
      [(not (struct? eq)) (error 'how-many-solutions "expected an equation as input")]
      [else
       (cond
         [(= 0 (equation-a eq)) "degenerate"]
         [(> (discriminant eq) 0) 2]
         [(= (discriminant eq) 0) 1]
         [else 0])])))

; TESTS
(= (how-many-solutions eq1) 2)
(= (how-many-solutions eq2) 1)
(= (how-many-solutions eq3) 0)
(string=? (how-many-solutions eq4) "degenerate")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; PROGRAMMING FUNDAMENTALS I - 2004
; MICHELE LANZA - FACULTY OF INFORMATICS - UNIVERSITY OF LUGANO, SWITZERLAND
; LECTURE 4 - COMPOUND DATA II AND SYNTAX AND SEMANTICS
; PERIMETER EXAMPLE

; DATA DEFINITION AND ANALYSIS
(define-struct square (origin side))
(define-struct circle (center radius))
; a shape is either
; 1. a square structure: (make-square o s) where 
; o is a posn and s is a number
; 2. a circle structure: (make-circle c r) where
; c is a posn and r is a number

; EXAMPLES
(define p1 (make-posn 50 50))
(define sq1 (make-square p1 30))
(define c1 (make-circle p1 40))

; TEMPLATE
; (define foo-shape
; (lambda (a-shape)
; (cond
; [(square? a-shape) ..(square-origin a-shape)..(square-side a-shape)..]
; [(circle? a-shape) ..(circle-center a-shape)..(circle-radius a-shape)..]

; perimeter: shape -> number
; computes the perimeter of a-shape
(define perimeter
  (lambda (a-shape)
    (cond
      [(square? a-shape) (* 4 (square-side a-shape))]
      [(circle? a-shape) (* 2 pi (circle-radius a-shape))]
      [else (error 'perimeter "wrong input given")])))

; TESTS
(= (perimeter sq1) 120)
(= (perimeter c1) (* 2 pi (circle-radius c1)))

; helper function perimeter-circle: circle -> number
(define perimeter-circle
  (lambda (a-circle)
    (* 2 pi (circle-radius a-circle))))
; helper function perimeter-square: square -> number
(define perimeter-square
  (lambda (a-square)
    (* 4 (square-side a-square))))

; perimeter2: shape -> number
; computes the perimeter of a-shape
(define perimeter2
  (lambda (a-shape)
    (cond
      [(square? a-shape) (perimeter-square a-shape)]
      [(circle? a-shape) (perimeter-circle a-shape)]
      [else (error 'perimeter "wrong input given")])))
