;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; PROGRAMMING FUNDAMENTALS I - 2004
; MICHELE LANZA - FACULTY OF INFORMATICS - UNIVERSITY OF LUGANO, SWITZERLAND
; LECTURE 10 - GENERATIVE RECURSION II AND ACCUMULATING KNOWLEDGE
; GRAPH TRAVERSAL

; a node is a symbol
; a graph is a (listof (list node (listof node)))

(define g1 '((A (B E)) (B (E F)) (C (D)) (D ()) (E (C F)) (F (D G)) (G ())))

; find-route : node node graph -> (listof node) or false
; creates a path from start-node to dest-node in a-graph 
; or produces false is there is no path
(define find-route 
  (lambda (start-node dest-node a-graph)
    (cond
      [(symbol=? start-node dest-node) (list dest-node)]
      [else (local (
                    (define potential-route 
                      (find-route-in-list (neighbors start-node a-graph) dest-node a-graph))
                    )
              (cond
                [(boolean? potential-route) false]
                [else (cons start-node potential-route)]))])))

; find-route-in-list : (listof node) node graph -> (listof node) or false
; creates a path from a node is lon to dest-node in a-graph
; or produces false
(define find-route-in-list
  (lambda (lon dest-node a-graph)
    (cond
      [(empty? lon) false]
      [else (local (
                    (define possible-route (find-route (first lon) dest-node a-graph))
                    )
              (cond
                [(boolean? possible-route) (find-route-in-list (rest lon) dest-node a-graph)]
                [else possible-route]))])))

; neighbors : node graph -> (listof node)
; produces the list of neighbors of a-node in a-graph
(define neighbors
  (lambda (a-node a-graph)
    (cond
      [(empty? a-graph) false]
      [else (cond
              [(eq? (first (first a-graph)) a-node) (second (first a-graph))]
              [else (neighbors a-node (rest a-graph))])])))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; PROGRAMMING FUNDAMENTALS I - 2004
; MICHELE LANZA - FACULTY OF INFORMATICS - UNIVERSITY OF LUGANO, SWITZERLAND
; LECTURE 10 - GENERATIVE RECURSION II AND ACCUMULATING KNOWLEDGE
; VECTOR EXAMPLE

; vector-sum : (vectorof number) -> number
; computes the sum of the numbers in a-vector
(define vector-sum
  (lambda (a-vector)
    (local (
            ; vector-sum-aux : (vectorof number) number -> number
            ; sums the numbers in a-vector with the index in [0,i)
            (define vector-aux-sum
              (lambda (a-vector i)
                (cond
                  [(= 0 i) 0]
                  [else (+ (vector-ref a-vector (- i 1))
                           (vector-aux-sum a-vector (- i 1)))])))
            )
      (vector-aux-sum a-vector (vector-length a-vector)))))

(define v1 (vector 2 3 5 7 11 13))

(= (vector-sum v1) 41)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; PROGRAMMING FUNDAMENTALS I - 2004
; MICHELE LANZA - FACULTY OF INFORMATICS - UNIVERSITY OF LUGANO, SWITZERLAND
; LECTURE 10 - GENERATIVE RECURSION II AND ACCUMULATING KNOWLEDGE
; ACCUMULATOR EXAMPLE 1

(define list1 (list 20 20 40 50 10))

; rel2abs : (listof number) -> (listof number)
; converts a list of relative distances to absolute distances
; the first item in the list is the distance to the origin
(define rel2abs
  (lambda (alon)
    (cond
      [(empty? alon) empty]
      [else (cons (first alon) (add-to-each (first alon) (rel2abs (rest alon))))])))

; add-to-each : number (listof number) -> (listof number)
; adds a-number to each item in alon
(define add-to-each
  (lambda (a-number alon)
    (cond
      [(empty? alon) empty]
      [else (cons (+ (first alon) a-number) (add-to-each a-number (rest alon)))])))

; rel2abs2 : (listof number) number -> (listof number)
; converts a list of relative distances to absolute distances using an accumulator
(define rel2abs2
  (lambda (alon current-sum)
    (cond
      [(empty? alon) empty]
      [else (cons (+ (first alon) current-sum) 
                  (rel2abs2 (rest alon) (+ (first alon) current-sum)))])))

; rel2abs3 : (listof number) -> (listof number)
; converts a list of relative distances to absolute distances using an accumulator
(define rel2abs3
  (lambda (alon)
    (local (
            (define rel2abstoo
              (lambda (alon accumulator)
                (cond
                  [(empty? alon) empty]
                  [else (cons (+ (first alon) accumulator)
                              (rel2abstoo (rest alon) (+ (first alon) accumulator)))])))
            )
      (rel2abstoo alon 0))))

; test
;(equal? (rel2abs list1) (rel2abs3 list1))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; PROGRAMMING FUNDAMENTALS I - 2004
; MICHELE LANZA - FACULTY OF INFORMATICS - UNIVERSITY OF LUGANO, SWITZERLAND
; LECTURE 10 - GENERATIVE RECURSION II AND ACCUMULATING KNOWLEDGE
; GRAPH TRAVERSAL WITH ACCUMULATOR

; a node is a symbol
; a pair is a list of two nodes: (cons n1 (cons n2 empty)) where n1 and n2 are symbols
; a simple-graph is a list of pairs : (listof pair)

; route-exists : node node graph -> boolean
; determines if there is a route from start-node to dest-node in a-graph
(define route-exists?
  (lambda (start-node dest-node a-graph)
    (cond
      [(symbol=? start-node dest-node) true]
      [else (route-exists? (neighbor start-node) dest-node a-graph)])))

; neighbor : node graph -> node
; determines the neighbor of a-node in a-graph
(define neighbor
  (lambda (a-node a-graph)
    (cond
      [(empty? a-graph) (error 'neighbor "no neighbor found!")]
      [else (cond
              [(symbol=? (first (first a-graph)) a-node) (second (first a-graph))]
              [else (neighbor a-node (rest a-graph))])])))

; route-exists?-acc : node node graph -> boolean
; determines if there is a route from start-node to dest-node in a-graph
(define route-exists?-acc
  (lambda (start-node dest-node a-graph)
    (local (
            ; re-accu? : node node graph (listof nodes) -> boolean
            ; determines if there is a route from start-node to dest-node 
            ; in a-graph using the accumulator visited-nodes
            (define re-accu?
              (lambda (start-node dest-node a-graph visited-nodes)
                (cond
                  [(symbol=? start-node dest-node) true]
                  [(contains? start-node visited-nodes) false]
                  [else (re-accu? (neighbor start-node a-graph) 
                                  dest-node 
                                  a-graph 
                                  (cons start-node visited-nodes))])))
            )
      (re-accu? start-node dest-node a-graph empty))))

; contains? : symbol (listof symbol) -> boolean
; determines if a symbol is present is a list of symbols
(define contains?
  (lambda (s alos)
    (cond
      [(empty? alos) false]
      [else (cond
              [(symbol=? s (first alos)) true]
              [else (contains? s (rest alos))])])))
      
               
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; PROGRAMMING FUNDAMENTALS I - 2004
; MICHELE LANZA - FACULTY OF INFORMATICS - UNIVERSITY OF LUGANO, SWITZERLAND
; LECTURE 10 - GENERATIVE RECURSION II AND ACCUMULATING KNOWLEDGE
; ACCUMULATOR EXAMPLES

; factorial1 : number -> number
; computes the factorial of a number
(define factorial1
  (lambda (n)
    (cond
      [(= 1 n) 1]
      [else (* n (factorial1 (- n 1)))])))

; factorial2 : number -> number
; computes the factorial of a number using an accumulator
(define factorial2
  (lambda (n)
    (local (
            (define factorial
              (lambda (n accumulator)
                (cond
                  [(= 1 n) accumulator]
                  [else (factorial (- n 1) (* accumulator n))])))
            )
      (factorial n 1))))

; invert1 : list -> list
; inverts a list
(define invert1
  (lambda (a-list)
    (cond
      [(empty? a-list) empty]
      [else (append (invert1 (rest a-list)) (list (first a-list)))])))

; invert2 : list -> list
; inverts a list
(define invert2
  (lambda (a-list)
    (cond
      [(empty? a-list) empty]
      [else (make-last (first a-list) (invert2 (rest a-list)))])))

(define make-last
  (lambda (an-item a-list)
    (cond
      [(empty? a-list) (list an-item)]
      [else (cons (first a-list) (make-last an-item (rest a-list)))])))

;(equal? (invert1 list1) (invert2 list1))

; invert3 : (listof x) -> (listof x)
; inverts a list using an accumulator
(define invert3
  (lambda (a-list)
    (local (
            (define reverse
              (lambda (a-list acc)
                (cond
                  [(empty? a-list) acc]
                  [else (reverse (rest a-list) (cons (first a-list) acc))])))
            )
      (reverse a-list empty))))

;(equal? (invert1 list1) (invert3 list1))
;(equal? (invert2 list1) (invert3 list1))

; fib-old : number -> number
; computes the xth fibonacci number
(define fib-old
  (lambda (x)
    (cond
      [(= x 1) 0]
      [(= x 2) 1]
      [else (+ (fib-old (- x 1)) (fib-old (- x 2)))])))

; fibonacci-acc : number -> number
; computes the xth fibonacci number using two accumulators
(define fibonacci-acc
  (lambda (n)
    (cond
      [(= n 0) 0]
      [else (local (
                    (define fib
                      (lambda (i acc1 acc2)
                        (cond
                          [(= i 1) acc1]
                          [else (fib (- i 1) (+ acc1 acc2) acc1)])))
                    )
              (fib n 0 1))])))

(= (fib-old 10) (fibonacci-acc 10))